/*
    *   This interface implements the easy communication through I2C-DEV 
    *
    *   Copyright (c)2019 Tóthpál István, istvan@tothpal.eu
    *
    *   TODO: use QT Debug instead of printf
*/

#include <linux/i2c-dev.h>
#include <i2c/smbus.h>

#include <stdio.h>                                                                                                                                           
#include <fcntl.h>                                                                                                                                           
#include <sys/ioctl.h>                                                                                                                                       
#include <unistd.h>                                                                                                                                          

#include "i2clinebus.h"
#include "../e2cmdw.h"

I2CLineBus::I2CLineBus() 
//: BusInterface()
{
	myI2CLine = 0;      //opened file that connects to i2c-x line 
	i2cbusnum = 4;		//interface setup will change to port you have choosen
	i2cdevaddr = 0x50;  //24c02 default address (it can be 0x50-0x57)              TODO:somewhere it could be choosen ( 24LC08B don't care A0-A2 pins and one of addressing bits, so it is mapped twice)
    clockdata = 0;
    dataout = 0;
    is_16bit_i2c_eprom = 0;
    memaddr_16bit.addrH = 0;
    memaddr_16bit.addrL = 0;
};

I2CLineBus::~I2CLineBus()
{
    if (myI2CLine!=0) {
	close( myI2CLine );
    };
};

void I2CLineBus::Set16bit(int is16bit) {
    is_16bit_i2c_eprom=is16bit;
};

int I2CLineBus::Open(int port) {
    char i2c_line_str[20];
    sprintf( i2c_line_str, "/dev/i2c-%d", port );

    // open i2c line
    if ( (myI2CLine = open( i2c_line_str, O_RDWR)) > 0 ) {
        printf("Open %s was succeeded!\n", i2c_line_str );
        i2cbusnum = port;
        Install( port );
        memaddr_16bit.addrH = 0;
        memaddr_16bit.addrL = 0;
        return OK;
    }
    else {
        printf("Error opening I2C line: %s\n", i2c_line_str );
	return -1;
    }
};

void I2CLineBus::Close() {
    if (myI2CLine > 0) {
        close( myI2CLine );
        myI2CLine = 0;
    };
    if (GetInstalled() >=0) {
        DeInstall();
    }
};

int I2CLineBus::TestOpen(int port) {
    char i2c_line_str[20];
    sprintf( i2c_line_str, "/dev/i2c-%d", port );

    // open i2c line
    if ( (myI2CLine = open( i2c_line_str, O_RDONLY)) > 0 ) {
        printf("Test open %s was succeeded!\n", i2c_line_str );
	return OK;
    }
    else {
        printf("Error opening I2C line: %s\n", i2c_line_str );
	return -1;
    }
};

int I2CLineBus::TestPort(int port) {
    int isok = 0;
    int result[128];

    if (myI2CLine <= 0 ) {
        isok = TestOpen( port );
    }
    if (isok == 0 ) {
        for (int i=0; i<128; i++) {
            __s32 opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, i );
            if (opResult==0) {
                result[i] = i2c_smbus_read_byte( myI2CLine );
            }
            else {
                result[i] = -1;
            }
        }
    
        // print result         TODO: it can be displayed at a popup window w/wo device selection
        printf("\n       00  01  02  03  04  05  06  07  08  09  0a  0b  0c  0d  0e  0f \n");
        for (int i=0; i<128; i++) {
            if ( (i%16) ==0 ) {
                printf(" %d0 : ", int(i/16));
            }
            printf( (result[i]>=0 ? " OK " : " -- ") );
            if ( ((i+1)%16) ==0 ) {
                printf("\n");
            }
        }

        Close();
//        isok = opResult;
    }
    return isok;
};

int I2CLineBus::TestSave(int port) {
    printf("TestSave port: %d not implemented!\n", port );
    return -1;
};

void I2CLineBus::TestRestore() {
    printf("TestRestore not implemented!\n");
};

int I2CLineBus::SetPower(bool onoff) {
    printf("SET Power: %s\n", onoff ? "true":"false");
    return 0;
};


int I2CLineBus::Read(int addr, uint8_t *data, long length, int page_size) {
    int isok = 0;
	__s32 err;
    __s32 opResult = 0;
    int result;
    
    if (myI2CLine <= 0 ) {
        isok = Open( i2cbusnum );
    }
    if (isok == 0 ) {
        if ((addr>=0) & (addr<128)) {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, addr );
            i2cdevaddr = addr;
        }
        else {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, i2cdevaddr );
        }
        isok = opResult;
    }
    if (isok == 0 ) {
        for ( long i=0; i<length; i=i+1 ) {
            if (is_16bit_i2c_eprom == 0) {
//                    printf("read:%d %d\n", i2cdevaddr, i);
//                    err = i2c_read_8bit( i2cdevaddr, i, (__u8*)(&result) );
                    err = i2c_smbus_read_byte_data( myI2CLine, i );
                    result = err;
            }
            else {
//                    printf("read:%d %d\n", memaddr_16bit.addrH, memaddr_16bit.addrL);
                err = i2c_read_16bit( i2cdevaddr, memaddr_16bit.addrH*256 + i, (__u8*)(&result) );
            }
            if (err >= 0) {
                *data++ = result;
            }
            else {
//                *data++ = i;
                return err;
            }
        };
    }
    return isok;
};

int I2CLineBus::i2c_read_16bit(__u8 slave_addr, __u16 memaddr, __u8 *result) {

    __u8 outbuf[2], inbuf[1];
    struct i2c_msg msgs[2];
    struct i2c_rdwr_ioctl_data msgset[1];

    outbuf[0] = int(memaddr/256);
    outbuf[1] = int(memaddr%256);

    msgs[0].addr = slave_addr;
    msgs[0].flags = 0;
    msgs[0].len = 2;
    msgs[0].buf = outbuf;

    msgs[1].addr = slave_addr;
    msgs[1].flags = I2C_M_RD | I2C_M_NOSTART;
    msgs[1].len = 1;
    msgs[1].buf = inbuf;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 2;

    inbuf[0] = 0;

    *result = 0;
    if (ioctl(myI2CLine, I2C_RDWR, &msgset) < 0) {
        perror("ioctl(I2C_RDWR) in i2c_read");
        return -1;
    }

    *result = inbuf[0];
    return 0;
}

int I2CLineBus::i2c_read_8bit(__u8 slave_addr, __u8 memaddr, __u8 *result) {

    __u8 outbuf[1], inbuf[1];
    struct i2c_msg msgs[2];
    struct i2c_rdwr_ioctl_data msgset[1];

    outbuf[0] = memaddr;

    msgs[0].addr = slave_addr;
    msgs[0].flags = 0;
    msgs[0].len = 1;
    msgs[0].buf = outbuf;

    msgs[1].addr = slave_addr;
    msgs[1].flags = I2C_M_RD | I2C_M_NOSTART;
    msgs[1].len = 1;
    msgs[1].buf = inbuf;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 2;

    inbuf[0] = 0;

    *result = 0;
    if (ioctl(myI2CLine, I2C_RDWR, &msgset) < 0) {
        perror("ioctl(I2C_RDWR) in i2c_read");
        return -1;
    }

    *result = inbuf[0];
    return 0;
}


int I2CLineBus::WriteWord(int addr, const uint8_t *data, long length, int page_size) {
    int isok = 0;
	int err;
    __s32 opResult = 0;
    
    if (myI2CLine <= 0 ) {
        isok = Open( i2cbusnum );
    }
    if (isok == 0 ) {
        if ((addr>=0) & (addr<128)) {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, addr );
        }
        else {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, i2cdevaddr );
        }
        isok = opResult;
    }
    if (isok == 0 ) {
        if (length==2) {
            err = i2c_smbus_write_byte_data( myI2CLine, (__u8)data[0] , (__u8)data[1] );
            if (err < 0) {
                return err;
            }
        };
        
    }
    return isok;
};

int I2CLineBus::WriteBlock(int addr, const uint8_t *data, long length, int page_size) {
    int isok = 0;
	int err;
    __s32 opResult = 0;
    char buf;
    
    if (myI2CLine <= 0 ) {
        isok = Open( i2cbusnum );
    }
    if (isok == 0 ) {
        if ((addr>=0) & (addr<128)) {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, addr );
            i2cdevaddr = addr;
        }
        else {
            opResult = ioctl(myI2CLine, I2C_SLAVE_FORCE, i2cdevaddr );
        }
        isok = opResult;
    }
    if (isok == 0 ) {
        if ((length>2) & (is_16bit_i2c_eprom==1 )){
            err = i2c_write_block_16bit( i2cdevaddr, (__u8*)data , length );
            if (err < 0) {
                return err;
            }
            else {
//                int retries = 0;
//                while (( read(myI2CLine,buf,1) <0 ) & (retries++<10));
                usleep( 10000 );
            }
        };
        
    }
    return isok;
};

int I2CLineBus::i2c_write_block_16bit(__u8 slave_addr, const __u8 *data, long length) {
    struct i2c_msg msgs[2];
    struct i2c_rdwr_ioctl_data msgset[1];

    msgs[0].addr = slave_addr;
    msgs[0].flags = 0;
    msgs[0].len = length;
    msgs[0].buf = (__u8 *)data;

    msgset[0].msgs = msgs;
    msgset[0].nmsgs = 1;

    if (ioctl(myI2CLine, I2C_RDWR, &msgset) < 0) {
        perror("ioctl(I2C_RDWR) in i2c_write-block");
        return -1;
    }

    return 0;
}

// used to write out the start address of reading/writeing, but my i2c_read_xxbit() will do it.
long I2CLineBus::StartWrite(uint8_t slave, const uint8_t *data, long length) {
    switch (length) {
        case 2:
//            printf("%d %d\n", data[0], data[1]);
            is_16bit_i2c_eprom = 1;
            memaddr_16bit.addrH = data[0];
            memaddr_16bit.addrL = data[1];
            break;                
        default:
            is_16bit_i2c_eprom = 0;
            memaddr_16bit.addrH = 0;
            memaddr_16bit.addrL = 0;
    };
    return length;
};


void I2CLineBus::SetControlLine(int res) {
    printf("Set Control Line: %d\n", res);
};

void I2CLineBus::SetDataOut(int sda) {
    printf("Set DATA OUT: %d\n", sda);
    dataout=sda;
};

void I2CLineBus::SetInvDataOut(int sda) {
    printf("Set INVDATA OUT: %d %d\n", sda, (sda ? 128 : 0) );
    dataout = (sda ? 128 : 0 ); 
};

void I2CLineBus::SetClock(int scl) {
    printf("SET Clock: %d\n", scl);
    clockdata=scl;
};

int I2CLineBus::GetDataIn() {
    printf("Getting DATA IN: %d\n", (dataout>0));
    return (dataout>0);
};

int I2CLineBus::GetClock() {
    printf("Getting CLOCK: %d\n", clockdata);
    return clockdata;
};

void I2CLineBus::SetClockData() {
    printf("SET Clock Data:\n");
    clockdata=1;
};

void I2CLineBus::ClearClockData() {
    printf("CLEAR Clock Data:\n");
    clockdata=1;
};

int I2CLineBus::IsClockDataUP() {
    printf("IS Clock Data UP:\n");
    return clockdata;
};

int I2CLineBus::IsClockDataDOWN() {
    printf("IS Clock Data DOWN:\n");
    return (clockdata==0);
};
